"""Bridges are how Schism services communicate with each other in a distributed context. They are never used when the
application is running in a monolithic mode. They are made of three parts:
- Bridge class: referenced in the schism.config, used to create instances of the bridge server and bridge client
- Server class: stands in front of the service class and accepts method call requests from clients
- Client class: used in place of the service class to proxy method calls as requests to the server

The client and server should be "dumb" in that they just pass payloads of dictionaries back and forth. There should be
no evaluation of how the results are handled. The active Schism controller handles everything including determining
where exceptions should be handled.

The bridge class should provide three different class methods:
- create_client: should return an instance of the bridge client
- create_server: should return an instance of the bridge server and register any launch tasks or endpoints that are
needed to make it accessible
- config_factory: should process any settings that can be passed to the bridge from a service config in the
schism.config file, this return in passed through to the create_client and create_server methods

The client class only has to implement a single method call_async_method which should pass the method call payload to
the bridge server. It should then return the result payload that the server responds with.

The server class's implementation is much less strict. It only needs to pass the method call payload it receives to the
call_async_method method of the server instance and then pass that method's return payload back to the client.

The client and server classes are instantiated with the bridge config, the server class is also instantiated with a
service facade that method call payloads can be passed to for handling."""
import traceback
from abc import ABC, abstractmethod
from functools import partial
from typing import Awaitable, Type, TYPE_CHECKING, Any, TypedDict

from bevy import get_repository

import schism.middleware as middleware

if TYPE_CHECKING:
    from schism.services import Service


class ResponseBuilder:
    """A helper context that captures all exceptions that are raised on the server while attempting to respond to a
    client request.

    In a standard single process application, errors raised by a method are handled by the caller. So it is necessary to
    capture all exceptions raised by a service and propagate them to the client so that the calling code can handle
    those exceptions as normal."""
    def __init__(self):
        self.payload: ResultPayload = ReturnPayload(result=None)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is not None:
            self.payload = ExceptionPayload(
                error=exc_val,
                traceback=traceback.format_exception(exc_type, exc_val, exc_tb),
            )

        return True

    def set(self, data):
        """Set the response payload. This payload is overwritten by any exceptions that are raised."""
        self.payload = ReturnPayload(result=data)


class RemoteError(Exception):
    """RemoteError is used to wrap a stacktrace passed from a service to a client. This error type is never raised on
    its own, instead it is given as the cause for an exception that is raised by the client to propagate an error that
    occurred on the server."""


class MethodCallPayload(TypedDict):
    service: "Type[Service]"
    method: str
    args: tuple
    kwargs: dict


class ReturnPayload(TypedDict):
    result: Any


class ExceptionPayload(TypedDict):
    error: Exception
    traceback: list[str]


type ResultPayload = ReturnPayload | ExceptionPayload


class BridgeClient(ABC):
    """Bridge clients proxy method calls from the service facade to the service server."""
    def __init__(self, config: Any):
        self.config = config

    @abstractmethod
    async def call_async_method(self, payload: MethodCallPayload) -> ResultPayload:
        """Should pass the method call payload up to the running bridge server and return the result payload it responds
        with."""
        ...


class BridgeServer:
    """Bridge servers take method call payloads and propagate them to the service itself, responding to the client with
    the result payload return from call_service."""
    def __init__(self, config: Any, service_facade: "BridgeServiceFacade"):
        self.config = config
        self.service_facade = service_facade

    def call_async_method(self, payload: MethodCallPayload) -> Awaitable[ResultPayload]:
        return self.service_facade.call_async_method(payload)



class BaseBridge(ABC):
    """Bridges provide methods for creating the corresponding configs, clients, and servers."""
    @classmethod
    @abstractmethod
    def create_client(cls, config: Any) -> BridgeClient:
        ...

    @classmethod
    @abstractmethod
    def create_server(cls, config: Any, service_facade: "BridgeServiceFacade") -> BridgeServer:
        ...

    @classmethod
    def config_factory(cls, bridge_config: Any) -> Any:
        return bridge_config


class BridgeClientFacade:
    """The client facade is injected in place of a service and passes off method calls to the bridge client. The facade
    handles propagation of exceptions from the bridge server to the client code. The facade also handles running
    middleware on the client side."""
    def __init__(
        self,
        bridge_type: Type[BaseBridge],
        service_type: "Type[Service]",
        config: Any,
        middleware_stack: "middleware.MiddlewareStackBuilder"
    ):
        self.client = bridge_type.create_client(config)
        self.service_type = service_type
        self.middleware = middleware_stack

    def __getattr__(self, item):
        return partial(self._call, item)

    async def _call(self, method: str, *args, **kwargs):
        middleware_stack = self.middleware.get_middleware(
            middleware.FilterEvent.CLIENT_CALL, middleware.FilterEvent.CLIENT_RESULT
        )

        payload = MethodCallPayload(
            service=self.service_type,
            method=method,
            args=args,
            kwargs=kwargs,
        )
        filtered_payload = await middleware_stack.filter(middleware.FilterEvent.CLIENT_CALL, payload)
        result = await self.client.call_async_method(filtered_payload)
        filtered_result = await middleware_stack.filter(middleware.FilterEvent.CLIENT_RESULT, result)
        return await self._process_result(filtered_result)

    async def _process_result(self, result: ResultPayload):
        match result:
            case {"error": error, "traceback": traceback}:
                raise error from RemoteError(
                    f"\n"
                    f"{''.join(traceback)}\n"
                    f"---------------------------------------------\n"
                    f"The above stacktrace is from a remote service\n"
                    f"---------------------------------------------"
                )

            case {"result": result}:
                return result

            case payload:
                raise RuntimeError(f"Impossible State, server response must be malformed: {payload!r}")


class BridgeServiceFacade:
    """The service facade gets the method call payload from the bridge server and handles calling the method on the
    service, capturing the return value and any exceptions to pass back to the bridge server as a result payload which
    is then sent to the client."""
    def __init__(self, service_type: "Type[Service]", middleware_stack: "middleware.MiddlewareStackBuilder"):
        self.service_type = service_type
        self.middleware = middleware_stack

    async def call_async_method(self, payload: MethodCallPayload) -> ResultPayload:
        middleware_stack = self.middleware.get_middleware(
            middleware.FilterEvent.SERVER_CALL, middleware.FilterEvent.SERVER_RESULT
        )
        with ResponseBuilder() as result:
            filtered_payload = await middleware_stack.filter(middleware.FilterEvent.SERVER_CALL, payload)

            if payload["service"] != self.service_type:
                raise RuntimeError(f"Service types do not match: {self.service_type} != {payload['service']}")

            service = get_repository().get(self.service_type)
            method = getattr(service, filtered_payload["method"])
            return_value = await method(*filtered_payload["args"], **filtered_payload["kwargs"])

            filtered_return = await middleware_stack.filter(middleware.FilterEvent.SERVER_RESULT, return_value)

            result.set(filtered_return)

        return result.payload



"""This module is intended solely to enable the schism command script that is configured in pyproject.toml."""
import os
import sys


def run():
    sys.path.append(os.getcwd())

    import schism.run
    schism.run.main(sys.argv[1:])




"""Schism controllers manage the state and life cycle of the application, storing the list of configured services,
active & remote services, entry points for active services, and tasks that need to be launched at runtime.

Schism app's have two basic life cycles: service and application.
- Service life cycle: Create distributed controller, bootstrap service types by instantiating their types, set up the
service entry points by loading them into the "schism.run" module namespace, and finally launching the service tasks in
the event loop.
- Application life cycle: Import the application module & access the application callback, create the distributed
controller with the application callback, and finally run the application callback as a launch task in the event loop.

Included are the MonolithicController and DistributedController.

The MonolithicController is intended to run the application entirely standalone, in a singular process. To ensure
everything runs consistent with the life cycle of the DistributedController, it is important to launch your application
using the "start_app" function. Here's a stripped down example:

    from schism import start_app
    ...
    async def main():
        ...

    if __name__ == '__main__':
        start_app(main())

This starts the "main" app function in an event loop that exits when the "main" function returns, and it handles
bootstrapping all service types so they start in the correct order following the correct life cycle.

The DistributedController handles bootstrapping standalone services and running the application callback as its own
process that is autowired to access distributed services. This is typically done by running services and applications
using the "schism run" CLI."""

import asyncio
from abc import ABC, abstractmethod
from typing import Any, Generator, Type, Callable, Awaitable

import bevy
from bevy import inject, dependency
from tramp.optionals import Optional

import schism.services as services
import schism.configs as configs
from schism.bridges import BridgeServiceFacade
from schism.middleware import MiddlewareStackBuilder

type ServicesConfigMapping = dict[Type[services.Service], configs.ServiceConfig]

_global_controller = None


def _validate_entry_point_name(name: str):
    if not name.isidentifier():
        raise RuntimeError(f"Entry point names must be valid Python identifiers: {name!r} (invalid)")

    if name.startswith("_"):
        raise RuntimeError(f"Entry point names cannot start with '_': {name!r} (invalid)")


class SchismController(ABC):
    def __init__(self, service: str = ""):
        self._service_configs: "Optional[dict[str, configs.ServiceConfig]]" = Optional.Nothing()
        self._active_services: Optional[ServicesConfigMapping] = Optional.Nothing()
        self._remote_services: Optional[ServicesConfigMapping] = Optional.Nothing()
        self._entry_points: dict[str, Any] = {}
        self._launch_tasks: list[Awaitable[None]] = []

    @property
    @abstractmethod
    def active_services(self) -> ServicesConfigMapping:
        """Returns a mapping of services that are running in the current process."""

    @property
    @abstractmethod
    def remote_services(self) -> ServicesConfigMapping:
        """Returns a mapping of services that are not running in the current process."""

    @abstractmethod
    def bootstrap(self):
        """Bootstraps the running process to make services available."""

    @property
    def entry_points(self) -> dict[str, Any]:
        return self._entry_points

    @property
    def service_configs(self) -> dict[str, configs.ServiceConfig]:
        match self._service_configs:
            case Optional.Some(service_configs):
                return service_configs

            case Optional.Nothing():
                self._service_configs = Optional.Some(
                    dict(self._load_services_configs())
                )
                return self.service_configs

    def add_launch_task(self, task: Awaitable[None]):
        self._launch_tasks.append(task)

    def create_entry_point(self, name: str, entry_point: Any):
        _validate_entry_point_name(name)
        self._entry_points[name] = entry_point

    def filter_services(
        self, condition: "Callable[[configs.ServiceConfig], bool]"
    ) -> "Generator[tuple[Type[services.Service], configs.ServiceConfig], None, None]":
        for service, service_config in self.service_configs.items():
            if condition(service_config):
                yield service_config.get_service_type(), service_config

    def find_service_matching(self, service: "Type[services.Service]") -> "Optional[configs.ServiceConfig]":
        for s in self.service_configs.values():
            if issubclass(service, s.get_service_type()):
                return Optional.Some(s)

        return Optional.Nothing()

    def get_service_config(self, service: "Type[services.Service]") -> "configs.ServiceConfig":
        match self.find_service_matching(service):
            case Optional.Some(service_config):
                return service_config

            case Optional.Nothing():
                raise ValueError(f"No service matching {service} is configured.")

    def is_service_active(self, service: "Type[services.Service]") -> bool:
        return self.get_service_config(service).get_service_type() in self.active_services

    def launch(self):
        if not self._launch_tasks:
            return

        asyncio.run(self._run_tasks())

    @inject
    def _load_services_configs(
       self, config: "configs.ApplicationConfig" = dependency()
    ) -> "Generator[tuple[str, configs.ServiceConfig], None, None]":
        for service_config in config.services:
            yield service_config.service, service_config

    async def _run_tasks(self):
        async with asyncio.TaskGroup() as group:
            for task in self._launch_tasks:
                await group.create_task(task)

    @classmethod
    def activate[Controller: SchismController](cls: Type[Controller], service: str = "") -> Controller:
        controller = cls(service)
        set_controller(controller)
        return controller

    @classmethod
    def start_application(cls, app: Awaitable[None]):
        """Activates the controller and starts the application."""
        controller = cls.activate()
        controller.add_launch_task(app)
        controller.launch()


class MonolithicController(SchismController):
    @property
    def active_services(self) -> ServicesConfigMapping:
        match self._active_services:
            case Optional.Some(active_services):
                return active_services

            case Optional.Nothing():
                self._active_services = Optional.Some(
                    {
                        service.get_service_type(): service
                        for service in self.service_configs.values()
                    }
                )
                return self.active_services

    @property
    def remote_services(self) -> ServicesConfigMapping:
        return {}

    def bootstrap(self):
        """Create all services."""
        for service_config in self.active_services.values():
            bevy.get_repository().get(service_config.get_service_type())  # Create the service instance

    def launch(self):
        self.bootstrap()
        super().launch()


class DistributedController(SchismController):
    def __init__(self, active_service: str):
        super().__init__()
        self._active_service_name = active_service
        self._servers = {}

    @property
    def active_services(self) -> ServicesConfigMapping:
        match self._active_services:
            case Optional.Some(active_services):
                return active_services

            case Optional.Nothing():
                self._active_services = Optional.Some(
                    dict(self.filter_services(lambda s: s.name == self._active_service_name))
                )
                return self.active_services

            case invalid_state:
                raise ValueError(f"Invalid state: {invalid_state}")

    @property
    def remote_services(self) -> ServicesConfigMapping:
        match self._remote_services:
            case Optional.Some(remote_services):
                return remote_services

            case Optional.Nothing():
                self._remote_services = Optional.Some(
                    dict(
                        self.filter_services(lambda s: s.service != self._active_service_name)
                    )
                )
                return self.remote_services

    def bootstrap(self):
        """Entry point processes need to bootstrap services that are active."""
        if not next(self.filter_services(lambda s: s.name == self._active_service_name), False):
            raise RuntimeError(
                f"Unknown service: {self._active_service_name}\n\nAll services must be configured in the "
                f"schism.config file."
            )

        for service_config in self.active_services.values():
            bevy.get_repository().get(service_config.get_service_type())  # Create the service instance
            self._launch_server(service_config)

    def _launch_server(self, service_config: configs.ServiceConfig):
        bridge = service_config.get_bridge_type()
        service_facade = BridgeServiceFacade(
            service_config.get_service_type(),
            MiddlewareStackBuilder(service_config.get_bridge_middleware()),
        )
        self._servers[service_config.service] = bridge.create_server(
            bridge.config_factory(service_config.bridge),
            service_facade,
        )


def get_controller() -> SchismController:
    global _global_controller

    if _global_controller is None:
        _global_controller = MonolithicController()

    return _global_controller


def has_controller() -> bool:
    return _global_controller is not None


def set_controller(controller: SchismController):
    global _global_controller

    _global_controller = controller


def start_app(app_callback: Awaitable[None]):
    MonolithicController.start_application(app_callback)



"""The Simple TCP Bridge provides a somewhat secure TCP client and server implementation for calling methods of a
service. Parameters, return values, and exceptions are passed using pickles and are secured by signing each payload
using a SHA256 hash salted with a secret key.

The secret key can be changed by setting the SCHISM_TCP_BRIDGE_SECRET environment variable. This should provide some
assurance that the pickles being passed are safe.

In the schism.config file it is possible to set the host and port that a service's TCP bridge server should be exposed
on and that the client should connect to.

Here's an example yaml config:

    services:
      - name: example
        service: example:Example
        bridge:
          type: schism.ext.bridges.simple_tcp:SimpleTCP
          serve_on: 0.0.0.0:1234

It is also possible to configure the client separately:

    services:
      - name: example
        service: example:Example
        bridge:
          type: schism.ext.bridges.simple_tcp:SimpleTCP
          serve_on: 0.0.0.0:1234
          client: example.com:4321


The Simple TCP Bridge uses a custom protocol on top of TCP. The version 0 protocol uses the following structure:

Usage          | Size (Bytes)   | Data Type
---------------|----------------|--------------------
Version        |              2 | int (big endian)
Content Length |              4 | int (big endian)
Signature      |             64 | bytes (sha256 hash)
Content        |      Arbitrary | pickle
---------------|----------------|--------------------
"""
import asyncio
import contextlib
import hashlib
import os
import pickle
from asyncio import StreamReader, StreamWriter
from functools import lru_cache

from schism.bridges import BaseBridge, BridgeClient, BridgeServer, BridgeServiceFacade, MethodCallPayload, ResultPayload
from schism.configs import SchismConfigModel
from schism.controllers import get_controller
from schism.middleware import MiddlewareStackBuilder


SIMPLE_TCP_VERSION_SUPPORTED = 0


def _generate_signature(data: bytes) -> bytes:
    return hashlib.sha256(data + SimpleTCP.SECRET_KEY).hexdigest().encode()


class SimpleTCPConfig(SchismConfigModel, lax=True):
    serve_on: str
    client: str


async def connect(host: str, port: int) -> tuple[StreamReader, StreamWriter]:
    try:
        return await asyncio.open_connection(host, port)
    except OSError as e:
        raise RuntimeError(f"Unable to connect to service on {host}:{port}") from e


async def read_version(reader: StreamReader) -> int:
    """Reads 2 bytes and converts them to a big endian int."""
    version = await reader.read(2)
    return int.from_bytes(version, byteorder="big")


async def read(reader: StreamReader) -> ResultPayload | MethodCallPayload:
    """When reading from a TCP connection first read 2 bytes to get the version, then 4 bytes to get the content length.
    Next read the 64 byte signature. Next read the content and validate the signature matches. If it does then it is
    safe to load the payload pickle.
    """
    version = await read_version(reader)
    if version != SIMPLE_TCP_VERSION_SUPPORTED:
        raise RuntimeError(
            f"Only SimpleTCP protocol version {SIMPLE_TCP_VERSION_SUPPORTED} is supported, detected "
            f"version {version}."
        )

    length_bytes = await reader.read(4)
    signature = await reader.read(64)

    length = int.from_bytes(length_bytes, byteorder="big")
    payload = await reader.read(length)
    if signature != _generate_signature(payload):
        raise ValueError(f"Received an invalid signature")

    return pickle.loads(payload)


async def send(data: ResultPayload | MethodCallPayload, writer: StreamWriter):
    """When writing to a TCP connection first write the 2 byte protocol version then the 4 byte content length of the
    pickled data, then the 64 byte signature, and finally write the pickle."""
    payload = pickle.dumps(data)
    writer.write(SIMPLE_TCP_VERSION_SUPPORTED.to_bytes(2, byteorder="big"))
    writer.write(len(payload).to_bytes(4, byteorder="big"))
    writer.write(_generate_signature(payload))
    writer.write(payload)
    await writer.drain()


class SimpleTCPClient(BridgeClient):
    config: SimpleTCPConfig

    @property
    @lru_cache
    def host(self) -> str:
        return (
            self.config.client
            if self.config.client
            else self.config.serve_on
        ).split(":")[0]

    @property
    def port(self) -> int:
        return int(
            (
                self.config.client
                if self.config.client
                else self.config.serve_on
            ).split(":")[1]
        )

    async def call_async_method(self, payload: MethodCallPayload):
        reader, writer = await connect(self.host, self.port)
        with contextlib.closing(writer):
            await send(payload, writer)
            return await read(reader)


class SimpleTCPServer(BridgeServer):
    config: SimpleTCPConfig

    @property
    @lru_cache
    def host(self) -> str:
        return self.config.serve_on.split(":")[0]

    @property
    @lru_cache
    def port(self) -> int:
        return int(self.config.serve_on.split(":")[1])

    async def launch(self):
        server = await asyncio.start_server(self._handle_request, self.host, self.port)

        async with server:
            await server.serve_forever()

    async def _handle_request(self, reader, writer):
        with contextlib.closing(writer):
            call_payload: MethodCallPayload = await read(reader)
            result = await self.call_async_method(call_payload)
            await send(result, writer)


class SimpleTCP(BaseBridge):
    SECRET_KEY = os.environ.get("SCHISM_TCP_BRIDGE_SECRET", "").encode()

    @classmethod
    def create_client(cls, config: SimpleTCPConfig) -> SimpleTCPClient:
        return SimpleTCPClient(config)

    @classmethod
    def create_server(cls, config: SimpleTCPConfig, service_facade: BridgeServiceFacade) -> SimpleTCPServer:
        server = SimpleTCPServer(config, service_facade)
        get_controller().add_launch_task(server.launch())
        return server

    @classmethod
    def config_factory(cls, bridge_config: str | dict[str, str | int]) -> SimpleTCPConfig:
        match bridge_config:
            case str() as serve_on:
                return SimpleTCPConfig(serve_on=serve_on, client=serve_on)

            case {"serve_on": str() as serve_on, "client": str() as client}:
                return SimpleTCPConfig(serve_on=serve_on, client=client)

            case {"serve_on": str() as serve_on}:
                return SimpleTCPConfig(serve_on=serve_on, client=serve_on)

            case _:
                raise ValueError(f"Invalid bridge configuration for {cls.__name__}: {bridge_config}")



import sys
import pathlib
from functools import lru_cache
from importlib import import_module
from typing import Type, TYPE_CHECKING, Any

from nubby import ConfigModel
from pydantic import BaseModel

from schism.middleware import Middleware

type StringOrSettings = str | dict[str, Any]

MAIN_MODULE_NAME = pathlib.Path(getattr(sys.modules["__main__"], "__file__", "")).stem or None

if TYPE_CHECKING:
    import schism.bridges as bridges
    import schism.services as services


class SchismConfigModel(BaseModel, ConfigModel, lax=True):
    """Base model config that implements the correct interface for serialization."""
    def to_dict(self):
        return self.model_dump()


class LaunchConfig(SchismConfigModel, lax=True):
    """Config model for settings related to launching an app with the "schism run" command. "app" should be a module
    import path and attribute name separated by a colon. "settings" is an optional dictionary of string keys that are
    valid Python identifiers with any kind of value, these are passed as keyword arguments to the callable attribute.

    Example:

        config = LaunchConfig(app="example:example_function", settings={"example": 1234})
    """
    app: str
    settings: dict[str, Any] | None = None


class ServiceConfig(SchismConfigModel, lax=True):
    """Config model for a service.
    - "name" is used for referencing the service in commands
    - "service" is the module import path and class name, separated by a colon, for the service class
    - "bridge" is either the module import path and class name, separated by a colon, for the bridge class, or a
    dictionary with a "type" key that is the bridge class string. All other keys in the dictionary are passed to the
    bridge types "config_factory" class method to generate teh config that is passed to the bridge client and server."""
    name: str
    service: str
    bridge: StringOrSettings

    def get_bridge_type(self) -> "Type[bridges.BaseBridge]":
        """Finds the module for the bridge type and gets the bridge type from the module."""
        return self._load_object(self._get_locator(self.bridge))

    def get_service_type(self) -> "Type[services.Service]":
        """Finds the module for the service type and gets the service type from the module."""
        return self._load_object(self.service)

    def get_bridge_config(self) -> Any:
        """Passes the bridge config to the bridge type's config factory method and passes back its return."""
        return self.get_bridge_type().config_factory(self.bridge)

    def get_bridge_middleware(self) -> dict[Type[Middleware], dict[str, Any]]:
        """Finds the module for the middleware type and gets the middleware type from the module."""
        match self.bridge:
            case {"middleware": list() as middleware}:
                return dict(self._generate_middleware(middleware))

            case _:
                return {}

    def _generate_middleware(self, middleware: list[StringOrSettings]):
        for middleware_setting in middleware:
            match middleware_setting:
                case str() as locator:
                    yield self._load_object(locator), {}

                case {"type": locator, **settings}:
                    yield self._load_object(locator), settings

                case _:
                    raise ValueError(f"Invalid configuration for service {self.name}: {middleware!r}")

    def _get_locator(self, string_or_settings: StringOrSettings) -> str:
        match string_or_settings:
            case str() as locator:
                return locator

            case {"type": str() as locator}:
                return locator

            case _:
                raise ValueError(f"Invalid configuration for service {self.name}: {string_or_settings!r}")

    @staticmethod
    @lru_cache
    def _load_object(import_path: str):
        module_path, attr = import_path.rsplit(":", 1)
        module = sys.modules["__main__"] if module_path == MAIN_MODULE_NAME else import_module(module_path)
        return getattr(module, attr)


class ApplicationConfig(SchismConfigModel, filename="schism.config"):
    """Config model for an application stored in the schism.config file. By default, this file can be a JSON, TOML, or
    YAML file. Schism only checks the working directory for this file."""
    services: list[ServiceConfig]
    launch: LaunchConfig | None = None



import os
import sys
from importlib import import_module
from typing import Any

from bevy import inject, dependency

from schism.configs import ApplicationConfig
from schism.controllers import SchismController, DistributedController


def start_services(service: str):
    controller = setup_controller(service)
    setup_entry_points(controller)
    controller.launch()


def setup_controller(service: str) -> SchismController:
    controller = DistributedController.activate(service)
    controller.bootstrap()
    return controller


def setup_entry_points(controller: SchismController):
    for name, entry_point in controller.entry_points.items():
        globals()[name] = entry_point


def start_application(module_path: str, application_callback_name: str, settings: dict[str, Any] | None = None):
    try:
        module = import_module(module_path)
    except ModuleNotFoundError as e:
        raise RuntimeError(f"The specified application module {module_path!r} could not be found.") from e

    try:
        application_callback = getattr(module, application_callback_name)
    except AttributeError as e:
        raise RuntimeError(
            f"The specified application callback {application_callback_name!r} could not be found in the "
            f"{module_path!r} module."
        ) from e

    DistributedController.start_application(application_callback(**settings or {}))


@inject
def start_application_using_config(config: ApplicationConfig = dependency()):
    if config.launch is None:
        raise RuntimeError("No launch config exists in schism.config")

    start_application(*config.launch.app.split(":"), settings=config.launch.settings)

def main(argv: list[str]):
    match argv:
        case ["run", "service", str() as service]:
            start_services(service)

        case ["run", str() as application_import] if ":" in application_import:
            start_application(*application_import.split(":"))

        case ["run"]:
            start_application_using_config()

        case _ if "SCHISM_ACTIVE_SERVICE" in os.environ:
            start_services(os.environ["SCHISM_ACTIVE_SERVICE"].strip())

        case _:
            print("""Welcome to Schism!

Schism is a simple service autowiring framework for Python. It allows you to write service oriented applications that
can also be easily be run as monoliths.

Usage:
    schism run service <service>        - Run the given service
    schism run <module>:<entry_point>   - Run the given application""")


if __name__ == "__main__":
    main(["run"] + sys.argv[1:])



"""Schism provides a middleware API that can be used to filter values going into and out of the client and server. There
are four filter events:
- client_call: Run when a call is made to a service method but before a request is made to the service server
- client_result: Run when the service server responds to a call request, right before the result is processed and passed
to the client code
- server_call: Run when the service server receives a call request, right before the service method is called
- server_result: Run when the service method returns, right before sending the response to the client

These middleware filters are only run on bridges and are never run when the application is being run in a monolithic
mode. They are solely intended to allow changes to be made to the payloads passing between services in a distributed
application.

To implement a middleware create a class inheriting from schism.middleware.Middleware and override the desired filter
methods.

    from schism.middleware import Middleware
    from schism.bridges import MethodCallPayload, ResultPayload, ReturnPayload

    class ExampleMiddleware(Middleware):
        async def filter_client_call(self, payload: MethodCallPayload) -> MethodCallPayload:
            ...

        async def filter_client_result(self, payload: ResultPayload) -> ResultPayload:
            ...

        def filter_server_call(self, payload: MethodCallPayload) -> MethodCallPayload:
            ...

        def filter_server_result(self, payload: ResultPayload) -> ResultPayload:
            ...

To apply a middleware to a bridge it must be added to the bridge config for a service in the schism.config file. Here's
an example YAML config:

    services:
      - name: example
        service: example:Example
        bridge:
          type: schism.ext.bridges.simple_tcp:SimpleTCP
          middleware:
            - example_middleware:ExampleMiddleware

It's possible to pass config settings into the middleware as keyword arguments:

    middleware:
      - type: example_middleware:ExampleMiddleware
        value: Example Value

The server and client lookup relevant middleware before running any filter events. They then instantiate the appropriate
middleware and run the necessary filters. This allows middleware instances to exist for the life span of each method
call. So a middleware with both server_call and server_result filters is instantiated before the server_call filter is
run and that same instance is used for both the server_call and server_result filter events.
"""
from enum import Enum
from typing import Awaitable, ClassVar, Type, Any

from schism.bridges import MethodCallPayload, ResultPayload, ReturnPayload


class FilterEvent(Enum):
    CLIENT_CALL = "client_call"  # Runs before the client sends a method call request
    CLIENT_RESULT = "client_result"  # Runs before the client processes the response
    SERVER_CALL = "server_call"  # Runs before the server processes the method call request
    SERVER_RESULT = "server_result"  # Runs before the server sends the response


class Middleware:
    __handlers__: ClassVar[set[FilterEvent]]

    def __init_subclass__(cls):
        # Find all filters provided by a middleware type
        cls.__handlers__ = {event for event in FilterEvent if f"filter_{event.value}" in vars(cls)}

    def __init__(self, **kwargs: Any):
        self.settings = kwargs

    def filter_client_call(self, payload: MethodCallPayload) -> Awaitable[MethodCallPayload]:
        """This filter is called before the client sends a method call request to a service."""
        raise NotImplementedError

    def filter_client_result(self, payload: ResultPayload) -> Awaitable[ResultPayload]:
        """This filter is called before the client processes the response from a service."""
        raise NotImplementedError

    def filter_server_call(self, payload: MethodCallPayload) -> Awaitable[MethodCallPayload]:
        """This filter is called before the server processes a method call request from a client."""
        raise NotImplementedError

    def filter_server_result(self, payload: ResultPayload) -> Awaitable[ResultPayload]:
        """This filter is called before the server sends a response to a client."""
        raise NotImplementedError


class MiddlewareStack:
    def __init__(self, middleware: list[Middleware]):
        self.middleware = middleware

    async def filter[T: MethodCallPayload | ReturnPayload](self, event: FilterEvent, payload: T) -> T:
        for middleware in self.middleware:
            if event in middleware.__handlers__:
                payload = await getattr(middleware, f"filter_{event.value}")(payload)

        return payload


class MiddlewareStackBuilder:
    def __init__(self, middleware: dict[Type[Middleware], dict[str, Any]]):
        self.middleware = middleware

    def get_middleware(self, *events: FilterEvent) -> MiddlewareStack:
        return MiddlewareStack(
            [
                middleware_type(**settings)
                for middleware_type, settings in self.middleware.items()
                if middleware_type.__handlers__.intersection(events)
            ]
        )



import schism.controllers
from schism.bridges import BridgeClientFacade
from schism.middleware import MiddlewareStackBuilder


class Service:
    @classmethod
    def __bevy_constructor__(cls):
        controller = schism.controllers.get_controller()

        # Direct injection for service that exist in the running process
        if controller.is_service_active(cls):
            return cls()

        # Inject a bridge client to remotely access a service that doesn't exist in the running process
        else:
            service_config = controller.get_service_config(cls)
            bridge = service_config.get_bridge_type()
            return BridgeClientFacade(
                bridge_type=bridge,
                service_type=cls,
                config=bridge.config_factory(service_config.bridge),
                middleware_stack=MiddlewareStackBuilder(service_config.get_bridge_middleware()),
            )
